@@ -40,6 +40,10 @@ class DefaultRequirements(SuiteRequirements):
             and not util.sqla_100
         )
 
+    @property
+    def check_constraints_w_enforcement(self):
+        return exclusions.fails_on("mysql")
+
     @property
     def unnamed_constraints(self):
         """constraints without names are supported."""
@@ -53,3 +57,10 @@ class DefaultRequirements(SuiteRequirements):
     @property
     def reflects_unique_constraints_unambiguously(self):
         return exclusions.fails_on("mysql")
+
+    @property
+    def reflects_pk_names(self):
+        """Target driver reflects the name of primary key constraints."""
+
+        return exclusions.fails_on_everything_except(
+            'postgresql', 'oracle', 'mssql', 'sybase')
